using System;

namespace MyNamespace
{
	/// <summary>
	///		SUBROUTINE CSVD - The singular value decomposition of a complex
	///		m by n matrix by ACM Algorithm 358
	/// </summary>

	public class Algo358
	{
		// no constructor logic

		public static void CSVD(ComplexF[,] a, int m, int n,
			int p, int nu, int nv, float[] s, ComplexF[,] u, ComplexF[,] v)
		// Singular Value Decomposition, a = u * s * Conjg(Tran(v)), a is destroyed by CSVD
		// the diagonal matrix s is output as a vector, m must be >= n
		// if smaller, a should be filled with zero rows
		// this code is adapted from Collected Algorithms from ACM, Algorithm 358
		// The transformation Conjg(Tran(u)) is applied to the p vectors given in columns
		// n, n+1, ..., n+p-1 of matrix a
		// See: http://www.scs.fsu.edu/~burkardt/f77_src/toms358/toms358.f
		// and: http://www.scs.fsu.edu/~burkardt/f77_src/toms358/toms358_prb.f
		{
			float[] b = new float[n];
			float[] c = new float[n];
			float cs, eps, eta, f, g, h;
			int i, j, k, k1, L, L1, nM1, np;
			ComplexF q = new ComplexF();
			ComplexF r = new ComplexF();
			float sn;
			float[] t = new float[n];
			float tol, w, x, y, z;
			eta = 1.5E-07F;			// eta = the relative machine precision
			tol = 1.5E-31F;			// tol = the smallest normalized positive number, divided by eta
			np = n + p;
			nM1 = n - 1;
			L = 0;
			//
			// HOUSEHOLDER REDUCTION
			c[0] = 0.0F;
			k = 0;
			while (true)
			{
				k1 = k + 1;
				//
				// ELIMINATION OF a(i, k), i = k, ..., m-1
				z = 0.0F;
				for (i = k; i < m; i++)
					z += a[i, k].r * a[i, k].r + a[i, k].i * a[i, k].i;
				b[k] = 0.0F;
				if (z > tol)
				{
					z = (float)Math.Sqrt(z);
					b[k] = z;
					w = ComplexF.Abs(a[k, k]);
					q &= ComplexF.Set(1.0F, 0.0F);
					if (w != 0.0F) q &= a[k, k] / w;
					a[k, k] &= q * (z + w);
					if (k != np - 1)
					{
						for (j = k1; j < np; j++)
						{
							q &= ComplexF.Set(0.0F, 0.0F);
							for (i = k; i < m; i++)
								q += ComplexF.Conjg(a[i, k]) * a[i, j];
							q /= z * (z + w);
							for (i = k; i < m; i++)
								a[i, j] -= q * a[i, k];
						}
					}
					//
					// PHASE TRANSFORMATION
					q &= -ComplexF.Conjg(a[k, k]) / ComplexF.Abs(a[k, k]);
					for (j = k1; j < np; j++)
						a[k, j] *= q;
				}
				//
				// ELIMINATION OF a(k, j], j = k+2, ..., n-1
				if (k == nM1) break;
				z = 0.0F;
				for (j = k1; j < n; j++)
					z += a[k, j].r * (float)a[k, j].r + a[k, j].i * a[k, j].i;
				c[k1] = 0.0F;
				if (z > tol)
				{
					z = (float)Math.Sqrt(z);
					c[k1] = z;
					w = ComplexF.Abs(a[k, k1]);
					q &= ComplexF.Set(1.0F, 0.0F);
					if (w != 0.0F) q &= a[k, k1] / w;
					a[k, k1] &= q * (z + w);
					for (i = k1; i < m; i++)
					{
						q &= ComplexF.Set(0.0F, 0.0F);
						for (j = k1; j < n; j++)
							q += ComplexF.Conjg(a[k, j]) * a[i, j];
						q /= z * (z + w);
						for (j = k1; j < n; j++) 
							a[i, j] -= q * a[k, j];
					}
					//
					// PHASE TRANSFORMATION
					q &= -ComplexF.Conjg(a[k, k1]) / ComplexF.Abs(a[k, k1]); 
					for (i = k1; i < m; i++)
						a[i, k1] *= q;
				}
				k = k1;
			}
			//
			// TOLERANCE FOR NEGLIGIBLE ELEMENTS
			eps = 0.0F;
			for (k = 0; k < n; k++)
			{
				s[k] = b[k];
				t[k] = c[k];
				if ((float)s[k] + t[k] > eps)
					eps = (float)s[k] + t[k];
			}
			eps *= eta;
			//
			//  INITIALIZATION OF u AND v
			if (nu > 0)
			{
				for (j = 0; j < nu; j++)
				{
					for (i = 0; i < m; i++)
						u[i, j] = ComplexF.Set(0.0F, 0.0F);
					u[j, j] = ComplexF.Set(1.0F, 0.0F);
				}
			}
			if (nv > 0)
			{
				for (j = 0; j < nv; j++)
				{
					for (i = 0; i < n; i++)
						v[i, j] = ComplexF.Set(0.0F, 0.0F);
					v[j, j] = ComplexF.Set(1.0F, 0.0F);
				}
			}
			//
			//  QR DIAGONALIZATION
			for (k = nM1; k >= 0; k--)
			{
				//
				// TEST FOR SPLIT
				while (true)
				{
					for (L = k; L >= 0; L--)
					{
						if (Math.Abs(t[L]) <= eps) goto Test;
						if (Math.Abs(s[L - 1]) <= eps) break;
					}
					//
					// CANCELLATION OF E(L)
					cs = 0.0F;
					sn = 1.0F;
					L1 = L - 1;
					for (i = L; i <= k; i++)
					{
						f = sn * t[i];
						t[i] *= cs;
						if (Math.Abs(f) <= eps) goto Test;
						h = s[i];
						w = (float)Math.Sqrt(f * f + h * h);
						s[i] = w;
						cs = h / w;
						sn = -f / w;
						if (nu > 0)
						{
							for (j = 0; j < n; j++)
							{
								x = u[j, L1].r;
								y = u[j, i].r;
								u[j, L1] = ComplexF.Set(x * cs + y * sn, 0.0F);
								u[j, i] = ComplexF.Set(y * cs - x * sn, 0.0F);
							}
						}
						if (np == n) continue;
						for (j = n; j < np; j++)
						{
							q &= a[L1, j];
							r &= a[i, j];
							a[L1, j] &= q * cs + r * sn;
							a[i, j] &= r * cs - q * sn;
						}
					}
					//
					// TEST FOR CONVERGENCE
			Test:	w = s[k];
					if (L == k) break;
					//
					// ORIGIN SHIFT
					x = s[L];
					y = s[k - 1];
					g = t[k - 1];
					h = t[k];
					f = ((y - w) * (y + w) + (g - h) * (g + h)) / (2.0F * h * y);
					g = (float)Math.Sqrt(f * f + 1.0F);
					if (f < 0.0F) g = -g;
					f = ((x - w) * (x + w) + (y / (f + g) - h) * h) / x;
					//
					// QR STEP
					cs = 1.0F;
					sn = 1.0F;
					L1 = L + 1;
					for (i = L1; i <= k; i++)
					{
						g = t[i];
						y = s[i];
						h = sn * g;
						g = cs * g;
						w = (float)Math.Sqrt(h * h + f * f);
						t[i - 1] = w;
						cs = f / w;
						sn = h / w;
						f = x * cs + g * sn;
						g = g * cs - x * sn;
						h = y * sn;
						y = y * cs;
						if (nv > 0)
						{
							for (j = 0; j < n; j++)
							{
								x = v[j, i - 1].r;
								w = v[j, i].r;
								v[j, i - 1] = ComplexF.Set(x * cs + w * sn, 0.0F);
								v[j, i] = ComplexF.Set(w * cs - x * sn, 0.0F);
							}
						}
						w = (float)Math.Sqrt(h * h + f * f);
						s[i - 1] = w;
						cs = f / w;
						sn = h / w;
						f = cs * g + sn * y;
						x = cs * y - sn * g;
						if (nu > 0)
						{
							for (j = 0; j < n; j++)
							{
								y = u[j, i - 1].r;
								w = u[j, i].r;
								u[j, i - 1] = ComplexF.Set(y * cs + w * sn, 0.0F);
								u[j, i] = ComplexF.Set(w * cs - y * sn, 0.0F);
							}
						}
						if (n == np) continue;
						for (j = n; j < np; j++)
						{
							q &= a[i - 1, j];
							r &= a[i, j];
							a[i - 1, j] &= q * cs + r * sn;
							a[i, j] &= r * cs - q * sn;
						}
					}
					t[L] = 0.0F;
					t[k] = f;
					s[k] = x;
				}
				//
				//  CONVERGENCE
				if (w >= 0.0F) continue;
				s[k] = -w;
				if (nv == 0) continue;
				for (j = 0; j < n; j++)
					v[j, k] = -v[j, k];
			}
			//
			//  SORT SINGULAR VALUES
			for (k = 0; k < n; k++)	// sort	descending	
			{
				g = -1.0F;
				j = k;
				for (i = k; i < n; i++)	// sort descending
				{
					if (s[i] <= g) continue;
					g = s[i];
					j = i;
				}
				if (j == k) continue;
				s[j] = s[k];
				s[k] = g;
				if (nv > 0)
				{
					for (i = 0; i < n; i++)
					{
						q &= v[i, j];
						v[i, j] &= v[i, k];
						v[i, k] &= q;
					}
				}
				if (nu > 0)
				{
					for (i = 0; i < n; i++)
					{
						q &= u[i, j];
						u[i, j] &= u[i, k];
						u[i, k] &= q;
					}
				}
				if (n == np) continue;
				for (i = n; i < np; i++)
				{
					q &= a[j, i];
					a[j, i] &= a[k, i];
					a[k, i] &= q;
				}
			}
			//
			//  BACK TRANSFORMATION
			if (nu > 0)
			{
				for (k = nM1; k >= 0; k--)
				{
					if (b[k] == 0.0F) continue;
					q &= -a[k, k] / ComplexF.Abs(a[k, k]);
					for (j = 0; j < nu; j++)
						u[k, j] *= q;
					for (j = 0; j < nu; j++)
					{
						q &= ComplexF.Set(0.0F, 0.0F);
						for (i = k; i < m; i++)
							q += ComplexF.Conjg(a[i, k]) * u[i, j];
						q /= ComplexF.Abs(a[k, k]) * b[k];
						for (i = k; i < m; i++)
							u[i, j] -= q * a[i ,k];
					}
				}
			}
			if (nv > 0 && n > 1)
			{
				for (k = n - 2; k >= 0; k--)
				{
					k1 = k + 1;
					if (c[k1] == 0.0F) continue;
					q &= -ComplexF.Conjg(a[k, k1]) / ComplexF.Abs(a[k, k1]);
					for (j = 0; j < nv; j++)
						v[k1, j] *= q;
					for (j = 0; j < nv; j++)
					{
						q &= ComplexF.Set(0.0F, 0.0F);
						for (i = k1; i < n; i++)
							q += a[k, i] * v[i, j];
						q &= q / (ComplexF.Abs(a[k, k1]) * c[k1]);
						for (i = k1; i < n; i++)
							v[i, j] -= q * ComplexF.Conjg(a[k, i]);
					}
				}
			}
		} // CSVD
	}
}

